/*
  User Interface with touch pad and lcd display
  BV4242
  Copyright (c) 2013 Jim Spence.  All right reserved.
  www.byvac.com - see terms and conditions for using hardware
  
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
  Release Notes:
  June 27 2014 PCB BV4242_b, Firmware 1.5  

*/  

#include "Arduino.h"
#include "Print.h"
#include "bv4243.h"

// =============================================================================
// SV3B Type serial control
// =============================================================================
// =============================================================================
// *****************************************************************************
// constructor associates i2c addresses
// *****************************************************************************
BV4243::BV4243(int8_t pad)
{
  _PADi2adr=pad;
  _displayMode=0x0c;
  _entryMode=0x06;
  _functionSet=0x38; // DL|N
  Wire.begin(); // join i2c bus
}

// =============================================================================
// Input
// =============================================================================
// =============================================================================

// **************************************************************
// gets a 8 bit value from the i2c bus
// **************************************************************
uint8_t BV4243::i2_8bit(int8_t i2adr)
{
uint8_t rv;
  Wire.requestFrom(i2adr, 1); // returns 1 byte
  rv = Wire.read(); // byte
//  Wire.endTransmission();
  return rv;
}

// **************************************************************
// gets a 16 bit value from the i2c bus
// **************************************************************
uint16_t BV4243::i2_16bit(int8_t i2adr)
{
uint16_t rv;
  Wire.requestFrom(i2adr, 2); // returns 2 bytes
  rv=Wire.read()*256; // high byte
  rv+=Wire.read(); // low byte
//  Wire.endTransmission();
  return rv;
}

// =============================================================================
// =============================================================================
// =============================================================================
// Keypad section
// =============================================================================
// *****************************************************************************
// send a key commands
// *****************************************************************************
void BV4243::sendKeyCmd(uint8_t cmd)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(cmd); // command
    Wire.endTransmission();
    delayMicroseconds(30); // probably not needed
}
// *****************************************************************************
// Clear keypad buffer
// *****************************************************************************
void BV4243::clrBuf()
{
    sendKeyCmd(1); // cl buf
    delayMicroseconds(250);
}

// *****************************************************************************
// Gets number of keys in buffer
// *****************************************************************************
uint8_t BV4243::keysBuf()
{
    sendKeyCmd(2); //
    return (i2_8bit(_PADi2adr));
}

// *****************************************************************************
// Gets a key returns 0 if no keys in buffer
// *****************************************************************************
uint8_t BV4243::key()
{
    sendKeyCmd(3); // command
    return (i2_8bit(_PADi2adr));
}

// *****************************************************************************
// See if a partiular key is in the buffer, returns 0 if not and it it is
// the position of the key within the buffer
// *****************************************************************************
uint8_t BV4243::keyIn(uint8_t k)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(4); // command
    Wire.write(k);
    Wire.endTransmission();
    return (i2_8bit(_PADi2adr));
}

// *****************************************************************************
// returns a key scan code 0 of no key pressed
// *****************************************************************************
uint8_t BV4243::scan()
{
    sendKeyCmd(5); // command
    return (i2_8bit(_PADi2adr));
}

// *****************************************************************************
// returns 8 average values in the buffer provided, values are 16 bits
// *****************************************************************************
void BV4243::avg(uint16_t *b)
{
uint16_t v;
    sendKeyCmd(10); // command
    Wire.requestFrom(_PADi2adr, 16); // returns 16 bytes
    for(uint8_t j=0;j < 16; j++) {
        v=Wire.read()*256; // high byte
        v+=Wire.read(); // low byte
        *(b++)=v;
    }
}

// *****************************************************************************
// returns 8 delta values in the buffer provided, values are 16 bits
// *****************************************************************************
void BV4243::delta(uint16_t *b)
{
uint16_t v;
    sendKeyCmd(11); // command
    Wire.requestFrom(_PADi2adr, 16); // returns 16 bytes
    for(uint8_t j=0;j < 16; j++) {
        v=Wire.read()*256; // high byte
        v+=Wire.read(); // low byte
        *(b++)=v;
    }
}

// *****************************************************************************
// Resets eeprom values
// *****************************************************************************
void BV4243::EEreset()
{
    sendKeyCmd(20); // command
}

// *****************************************************************************
// Sleep mode, the keypad will not function in this mode but the device can 
// be awakend vie a read or write to I2C
// *****************************************************************************
void BV4243::sleep()
{
    sendKeyCmd(21); // command
}

// *****************************************************************************
// rings beeper
// *****************************************************************************
void BV4243::beep(uint8_t beepMs)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(6); // command
    Wire.write(beepMs);
    Wire.endTransmission();
}

// =============================================================================
// EEPROM settings
// =============================================================================
// *****************************************************************************
// sets new trigger value
// *****************************************************************************
void BV4243::trigger(uint16_t value)
{
    EEwrite(EE_TRIGGERH,(value >> 8) & 0xff);
    EEwrite(EE_TRIGGERL,value & 0xff);
}

// *****************************************************************************
// sets new hysteresis value
// *****************************************************************************
void BV4243::hyst(uint8_t value)
{
    EEwrite(EE_HYST,value);
}

// *****************************************************************************
// sets new keytable pointer
// *****************************************************************************
void BV4243::keyPtr(uint8_t value)
{
    EEwrite(EE_KEYPTR,value);
}

// *****************************************************************************
// sets new key table size
// *****************************************************************************
void BV4243::keySize(uint8_t value)
{
    EEwrite(EE_KEYSIZE,value);
}

// *****************************************************************************
// sets new debounce
// *****************************************************************************
void BV4243::debounce(uint8_t value)
{
    EEwrite(EE_DEBOUNCE,value);
}

// *****************************************************************************
// sets new repeat value
// *****************************************************************************
void BV4243::repeat(uint16_t value)
{
    EEwrite(EE_REPEATH,(value >> 8) & 0xff);
    EEwrite(EE_REPEATL,value & 0xff);
}

// *****************************************************************************
// sets new timebase
// *****************************************************************************
void BV4243::timebase(uint8_t value)
{
    EEwrite(EE_TIMEBASE,value);
}

// *****************************************************************************
// sets new defailt bl
// *****************************************************************************
void BV4243::defaultBL(uint8_t value)
{
    EEwrite(EE_BL,value);
}

// =============================================================================
// SV3 system section
// =============================================================================
// *****************************************************************************
// Writes a byte to the eeprom at a given address
// *****************************************************************************
void BV4243::EEwrite(uint8_t adr, uint8_t value)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(0x91); // command
    Wire.write(adr);
    Wire.write(value);
    Wire.endTransmission();
}

// *****************************************************************************
// Reads a byte from eeprom at given address
// *****************************************************************************
uint8_t BV4243::EEread(uint8_t adr)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(0x90); // command
    Wire.write(adr);
    Wire.endTransmission();
    return (i2_8bit(_PADi2adr));
}

// *****************************************************************************
// Get device ID as a 16 bit number
// *****************************************************************************
uint16_t BV4243::ID()
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(0xa1); // command
    Wire.endTransmission();
    return (i2_16bit(_PADi2adr));
}

// *****************************************************************************
// Get firmware version as two bytes
// *****************************************************************************
void BV4243::Version(uint8_t *b)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(0xa1); // command
    Wire.endTransmission();
    Wire.requestFrom(_PADi2adr, 2); // returns 16 bytes
    *(b++)=Wire.read(); // high byte
    *(b)=Wire.read(); // low byte   
}

// =============================================================================
// =============================================================================
// =============================================================================
// LCD Section
// This is for the ST7032i controller and follows the same commands used 
// for LiquidCrystal; to this end column is specified first i.e 16x2
// =============================================================================
// 
// *****************************************************************************
// sends a command to the display, bus is opened and closed useful for
// single commands
// *****************************************************************************
void BV4243::cmd(uint8_t cmd)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(31);
    Wire.write(cmd);
    Wire.endTransmission();
}

// *****************************************************************************
// sends data to the display, bus open at this point
// *****************************************************************************
inline size_t BV4243::write(uint8_t data)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(32);
    Wire.write(data);
    Wire.endTransmission();
    return 1;
}

// *****************************************************************************
// Initialise display
// *****************************************************************************
void BV4243::lcd_reset()
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(30);
    Wire.endTransmission();
    delay(450);
}

// *****************************************************************************
// sets back light brightness 0 - 130 but setting higher does not matter
// *****************************************************************************
void BV4243::bl(uint8_t r, uint8_t g, uint8_t b)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(36); // command
    Wire.write(r); // byte
    Wire.write(g); // byte
    Wire.write(b); // byte
    Wire.endTransmission();
}

// *****************************************************************************
// contrast from 0 to 63 
// *****************************************************************************
void BV4243::lcd_contrast(uint8_t m)
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(37); // command
    Wire.write(m);
    Wire.endTransmission();
}

// *****************************************************************************
// Displays the current sign on message
// *****************************************************************************
void BV4243::lcd_startMsg()
{
    Wire.beginTransmission(_PADi2adr);
    Wire.write(35); // command
    Wire.endTransmission();
    delay(10);
}

// =============================================================================
// Matches LiquidCrystal Class
// *****************************************************************************
// 
// *****************************************************************************
void BV4243::clear()
{
    cmd(0x1); // cmd for clear
    delay(30); // extra big delay
}

void BV4243::home()
{
    cmd(0x2); // dbl ht
    delay(30); // big delay
}

void BV4243::setCursor(int8_t col, int8_t row)
{
uint8_t adr;
    if(--row < 0) row = 0;
    if(row > 1) row = 1;
    if(--col < 0) col = 0;
    if(col > 15) col = 15;
    if(row == 0 ) adr=0x80;
    else adr=0xc0;
    cmd(adr+col);
}

void BV4243::cursor()
{
    bitset(_displayMode,1);
    cmd(_displayMode);
}

void BV4243::noCursor()
{
    bitclr(_displayMode,1);
    cmd(_displayMode);
}

void BV4243::blink()
{
    bitset(_displayMode,0);
    cmd(_displayMode);
}

void BV4243::noBlink()
{
    bitclr(_displayMode,0);
    cmd(_displayMode);
}

void BV4243::display()
{
    bitset(_displayMode,2);
    cmd(_displayMode);
}

void BV4243::noDisplay()
{
    bitclr(_displayMode,2);
    cmd(_displayMode);
}

void BV4243::scrollDisplayLeft()
{
    cmd(0x18);
}

void BV4243::scrollDisplayRight()
{
    cmd(0x1c);
}

void BV4243::autoscroll()
{
    bitset(_entryMode,0);
    cmd(_entryMode);
}

void BV4243::noAutoscroll()
{
    bitclr(_entryMode,0);
    cmd(_entryMode); 
}

void BV4243::leftToRight()
{
    bitset(_entryMode,1);
    cmd(_entryMode); 
}

void BV4243::rightToLeft()
{
    bitclr(_entryMode,1);
    cmd(_entryMode); 
}

void BV4243::createChar(uint8_t location, uint8_t charmap[]) 
{
  location &= 0x7; // we only have 8 locations 0-7
  cmd(0x40 | (location << 3));
  delayMicroseconds(30);
  for (int i=0; i<8; i++) {
    write(charmap[i]);
    delayMicroseconds(30);
  }
}
